package turingmachine;

import java.awt.*;

import javax.swing.*;

/**
 * All components that get added to a view component need to inherit from this.
 */
abstract class CTMComponent
	extends JComponent
{
  /**
   * The components corresponding data instance.
   */
  private CTMData mData;

  /**
   * The modification of mData that this component is currently representing.
   */
  private int mModificationCount;

  /**
   * Constructor.
   *
   * @param aData The corresponding data instance for this component.
   */
  public CTMComponent( CTMData aData )
  {
	setData( aData );
  }

  public int getModificationCount()
  {
	return mModificationCount;
  }

  /**
   * Bring component to the front of its view.
   */
  public void toFront()
  {
	Container viewContainer = getParent();

	viewContainer.remove( this );
	viewContainer.add( this, 0 );
	viewContainer.repaint();
  }

  /**
   * Send component to the back of its view.
   */
  public void toBack()
  {
	Container viewContainer = getParent();

	viewContainer.remove( this );
	viewContainer.add( this, viewContainer.getComponentCount() );
	viewContainer.repaint();
  }

  /**
   * Set this components corresponding data instance.
   *
   * @param aData The corresponding data instance for this component.
   */
  public void setData( CTMData aData )
  {
	mModificationCount = -1;
	mData = aData;
  }

  /**
   * Gets the data instance for this component.
   *
   * @returns The data instance this components syncs to.
   */
  public CTMData getData()
  {
	return mData;
  }

  /**
   * Synchronise this component with its data instance.
   */
  public void sync()
  {
	mModificationCount = getData().getModificationCount();
	repaint();
  }

  /**
   * Works out whether this component is in sync with its data instance.
   *
   * @return true, if data and component modification counts are the same.<br>
   *         false, the modification counts are different.
   */
  public boolean requiresSync()
  {
	return getData().getModificationCount()!=mModificationCount;
  }

  /**
   * Get the view component this link component belongs to.
   *
   * @return The view component that this link component belongs to.
   */
  public CTMComponentView getViewComponent()
  {
	JComponent parent = this;

	while ( ( parent = ( JComponent ) parent.getParent() )!=null )
	{
	  if ( parent instanceof CTMComponentView )
	  {
		return ( CTMComponentView ) parent;
	  }
	}

	// should never get here
	throw new RuntimeException(
		"getViewComponent() component is not child of a CTMViewComponent" );

  }

  /**
   * Get the document that this state is from.
   *
   * @return The document this state is from.
   */
  public CTMDataDocument getDocument()
  {
	return getData().getDocument();
  }
}
