package turingmachine;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;

/**
 * This class manages a small draggable handle for moving the control point on
 * CTMComponentLinks.
 */
class CTMComponentControl
	extends JComponent
{
  /**
   * The link component that this control is currently controlling.
   */
  private CTMComponentLink mLinkComponent = null;

  /**
   * Mouse Pointer offset from the center of the component when dragged.
   */
  private Point mDragOffset = new Point();

  /**
   * Constructor
   */
  public CTMComponentControl()
  {
	addMouseMotionListener( new MouseMotionAdapter()
	{
	  public void mouseDragged( MouseEvent e )
	  {
		this_mouseDragged( e );
	  }
	} );

	addMouseListener( new MouseAdapter()
	{
	  public void mousePressed( MouseEvent e )
	  {
		requestFocus();
		this_mousePressed( e );
	  }

	  public void mouseReleased( MouseEvent e )
	  {
		this_mouseReleased( e );
	  }
	} );

	addKeyListener( new KeyAdapter()
	{
	  public void keyPressed( java.awt.event.KeyEvent e )
	  {
		if ( e.getKeyCode()==KeyEvent.VK_DELETE )
		{
		  mLinkComponent.getLinkData().delete();
		}
	  }
	} );

	setCursor( Cursor.getPredefinedCursor( Cursor.MOVE_CURSOR ) );
	setSize( 5, 5 );
  }

  /**
   * Sets the link component that this control is adjusting.
   *
   * @param aLinkComponent The component to assign to this control.
   */
  public void setLinkComponent( CTMComponentLink aLinkComponent )
  {
	mLinkComponent = aLinkComponent;
  }

  /**
   * Get the link component that this control is controlling.
   *
   * @return The link component that this component is controlling.
   */
  private CTMComponentLink getLinkComponent()
  {
	return mLinkComponent;
  }

  /**
   * Get the link data that this component is modifying.
   *
   * @return The link data that this component is modifying. If not set
   *         returns null.
   */
  private CTMDataLink getLinkData()
  {
	if ( mLinkComponent==null )
	{
	  return null;
	}
	return mLinkComponent.getLinkData();
  }

  /**
   * Get the view component that this control is belongs to.
   *
   * @return The view component that this component belongs to.
   */
  private CTMComponentView getViewComponent()
  {
	return ( CTMComponentView ) getParent();
  }

  /**
   * Get the document that this control is modifying.
   *
   * @return the document that this component is modifying.
   */
  private CTMDataDocument getDocument()
  {
	return getViewComponent().getDocument();
  }

  /**
   * Does this control control aLinkComponent?
   *
   * @param aLinkComponent The link component we want to test for control.
   * @return true aLinkComponent is being controlled by this control, else
   *         false.
   */
  public boolean isControlling( CTMComponentLink aLinkComponent )
  {
	return aLinkComponent==mLinkComponent;
  }

  /**
   * Paint this component.
   *
   * @param g The graphics context to paint this component on.
   */
  public void paint( Graphics g )
  {
	// handle is currently a simple black filled rectangle.
	g.setColor( getDocument().getPreferences().control_colour );
	g.fillRect( 0, 0, getWidth(), getHeight() );
  }

  /**
   * Handles the mouse being pressed on the control.
   *
   * @param e The mouse pressed event.
   */
  private void this_mousePressed( MouseEvent e )
  {
	mDragOffset.setLocation( e.getPoint() );

	// bring component to front
	/*Container j = getParent();
	  j.remove( this );
	  j.add( this, 0 );*/
  }

  /**
   * Handles the mouse dragging the control.
   *
   * @param e The dragging event.
   */
  private void this_mouseDragged( MouseEvent e )
  {
	// translate the event location into view coordinates
	e.translatePoint( getX()-mDragOffset.x, getY()-mDragOffset.y );

	// move this component to where the mouse dragged to
	setLocation( e.getPoint() );

	// translate point to link local coordinates using linkComponents affine
	// transform
	CTMComponentLink linkComponent = getLinkComponent();

	try
	{
	  Point srcPoint = getLocation();
	  srcPoint.x += getWidth()/2;
	  srcPoint.y += getHeight()/2;

	  getLinkData().setCtrl(
		  linkComponent.transform.inverseTransform( srcPoint,
		  new Point() ) );

	  getViewComponent().sync();
	}
	catch ( java.awt.geom.NoninvertibleTransformException ex )
	{
	  ex.printStackTrace();
	}

	getViewComponent().checkSize();
  }

  /**
   * Handle a mouse release event on the control.
   *
   * @param e The mouse released event.
   */
  private void this_mouseReleased( MouseEvent e )
  {
	getDocument().notifyObservers();
  }
}
