package turingmachine;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;

import javax.swing.*;

/**
 * Manages the link component that is draw between states.
 */
class CTMComponentLink
	extends CTMComponent
{
  /**
   * The curve that makes up half of the link.
   */
  private java.awt.geom.QuadCurve2D.Double curve1 = new java.awt.geom.QuadCurve2D.Double();

  /**
   * The curve that makes up half of the link.
   */
  private java.awt.geom.QuadCurve2D.Double curve2 = new java.awt.geom.QuadCurve2D.Double();

  /**
   * Arrows heads left hand edge on curve end.
   */
  private CTMPoint arrowLeft = new CTMPoint();

  /**
   * Arrows heads right hand edge on curve end.
   */
  private CTMPoint arrowRight = new CTMPoint();

  /**
   * Transform from local to view coords ( and back via inverse ).
   */
  public AffineTransform transform = new AffineTransform();

  /**
   * Transition component.
   */
  public JPanel panelTransitionComponents;

  /**
   * Constructor.
   *
   * @param linkData The link data this link component represents.
   */
  public CTMComponentLink( CTMDataLink linkData )
  {
	super( linkData );

	addMouseListener( new MouseAdapter()
	{
	  public void mouseClicked( MouseEvent e )
	  {
		// TODO: is this correct, I thought a mouse click should auto
		// focus
		requestFocus();
		this_mouseClicked( e );
	  }
	} );

	this.addFocusListener( new FocusListener()
	{
	  public void focusGained( FocusEvent e )
	  {
		showControl();
	  }

	  public void focusLost( FocusEvent e )
	  {
		// hide controller
		CTMComponentView v = getViewComponent();
		CTMComponentControl controlComponent = v.getControlComponent();
		//controlComponent.setVisible( false );
		// TODO
	  }
	} );

	this.setFocusable( true );

	jbInit();
  }

  /**
   * Initalise the components used for the link.
   */
  private void jbInit()
  {
	panelTransitionComponents = new JPanel();
	panelTransitionComponents.setLayout( new GridLayout( 0, 1 ) );
	panelTransitionComponents.setOpaque( false );
	add( panelTransitionComponents );
  }

  /**
   * Handle mouse clicked on link.
   *
   * @param e The mouse click event.
   */
  private void this_mouseClicked( MouseEvent e )
  {
	CTMComponentView v = getViewComponent();

	switch ( v.getSelectedTool() )
	{
	  // If the mouse is clicked on the link
	  // then show a control knob to adjust the curve.
	  case CREATE_LINK:
	  case HAND:
		showControl();
		break;

	  case DELETE:
		getLinkData().delete();

		//getDocument().notifyObservers();
		break;
	}
  }

  /**
   * Show the adjustment control on the curve.
   */
  public void showControl()
  {
	CTMComponentView v = getViewComponent();
	CTMComponentControl controlComponent = v.getControlComponent();
	// bring control to front
	v.remove( controlComponent );
	v.add( controlComponent, 0 );
	controlComponent.setLinkComponent( this );
	controlComponent.setVisible( true );
	updateControls();
  }

  /**
   * Catchs the link being removed from the view and updates the
   * controlComponent if need be.
   */
  public void removeNotify()
  {
	super.removeNotify();

	CTMComponentControl controlComponent = getViewComponent()
		.getControlComponent();
	if ( controlComponent.isControlling( this ) )
	{
	  controlComponent.setVisible( false );
	}
  }

  /**
   * Is the point x,y roughly on the link curve.
   *
   * @param x The x coordinate to test.
   * @param y The y coordinate to test.
   * @return true - it does contain this point<br>
   *         false - it does not contain the point
   */
  public boolean contains( int x, int y )
  {
	return ( curve1.intersects( x-1, y-1, 3, 3 )||curve2.intersects(
		x-1, y-1, 3, 3 ) )
		||panelTransitionComponents.contains( x
		-panelTransitionComponents.getX(), y
		-panelTransitionComponents.getY() );
  }

  /**
   * Paint link arrow.
   *
   * @param g The graphics context to paint it on.
   */
  public void paint( Graphics g )
  {
	super.paint( g );

	Graphics2D g2 = ( Graphics2D ) g;
	CTMPreferences prefs = getDocument().getPreferences();

	g2.setStroke( new BasicStroke( prefs.link_lineWidth ) );
	g2.setColor( prefs.link_lineColour );

	g2.draw( curve1 );
	g2.draw( curve2 );

	g2.setColor( prefs.link_arrowColour );

	if ( prefs.link_arrowType.compareToIgnoreCase( "filled" )==0 )
	{
	  g2.fillPolygon( new int[]
		  {
		  arrowLeft.x, ( int ) curve2.getX2(),
		  arrowRight.x}, new int[]
		  {
		  arrowLeft.y,
		  ( int ) curve2.getY2(), arrowRight.y}, 3 );
	}
	else
	{
	  g2.drawLine( arrowLeft.x, arrowLeft.y, ( int ) curve2.getX2(),
		  ( int ) curve2.getY2() );

	  g2.drawLine( ( int ) curve2.getX2(), ( int ) curve2.getY2(),
		  arrowRight.x, arrowRight.y );
	}
  }

  /**
   * Synchronize this link with it's data component.
   */
  public void sync()
  {
	if ( requiresSync() )
	{
	  for ( Iterator k = getLinkData().transitionsIterator(); k.hasNext(); )
	  {
		CTMDataTransition transitionData = ( CTMDataTransition ) k.next();
		CTMComponentTransition comp = ( CTMComponentTransition ) getComponentByData( transitionData );

		if ( comp==null&&!transitionData.isDeleted() )
		{
		  comp = new CTMComponentTransition( transitionData );

		  panelTransitionComponents.add( comp );

		  if ( getViewComponent().isSelected() )
		  {
			comp.select();
		  }
		}

		// sync component to its data.
		if ( !transitionData.isDeleted() )
		{
		  comp.sync();
		}
	  }

	  getViewComponent().syncComponents( panelTransitionComponents );

	  panelTransitionComponents.setSize( panelTransitionComponents
		  .getPreferredSize() );
	  setCurve();
	  updateControls();
	  panelTransitionComponents.validate();
	  validate();

	  super.sync();
	}
  }

  /**
   * Get the Component that data is referred to in.
   *
   * @param aData The data the defines a components parameters.
   * @return The component if found, else null.
   */
  public CTMComponent getComponentByData( CTMData aData )
  {
	// loop through all the views components looking for a component of type
	// CTMComponent that has a data the same as aData.
	Component c[] = ( Component[] ) panelTransitionComponents.getComponents();

	for ( int cnt = 0; cnt<c.length; cnt++ )
	{
	  if ( c[cnt] instanceof CTMComponent )
	  {
		CTMComponent component = ( CTMComponent ) c[cnt];
		if ( component.getData()==aData )
		{
		  return component;
		}
	  }
	}

	return null;
  }

  /**
   * Update the curve controls, if the curve is focus.
   */
  private void updateControls()
  {
	CTMPoint tempPoint = new CTMPoint();

	CTMComponentControl controlComponent = getViewComponent().getControlComponent();

	if ( controlComponent.isControlling( this ) )
	{
	  // transform the control point from local coordinates to view
	  // coordinates
	  transform.transform( getLinkData().getCtrl(), tempPoint );

	  // make the center of the control sit on the curve
	  tempPoint.subtract( controlComponent.getWidth()/2,
		  controlComponent.getHeight()/2 );

	  // update the controls location
	  controlComponent.setLocation( tempPoint );
	}
  }

  /**
   * Set the curves parametrics depending on the location of the states it
   * links. + controlPoint + controlPoint1 + controlPoint2 + startpoint +
   * endPoint A bezier is draw between these points. controlPoint1 =
   * midPoint(endPoint->controlPoint) + (midPoint(endPoint->controlPoint)
   * rotated 90degrees)/2 controlPoint2 = midPoint(endPoint->controlPoint) +
   * (midPoint(endPoint->controlPoint) anti-rotated 90degrees)/2
   */
  private void setCurve()
  {
	CTMPoint controlPoint = new CTMPoint();
	CTMPoint controlPoint1 = new CTMPoint();
	CTMPoint controlPoint2 = new CTMPoint();

	AffineTransform mTempTransform = new AffineTransform();

	CTMDataLink linkData = getLinkData();
	CTMDataState stateFrom = linkData.getStateFrom();
	CTMDataState stateTo = linkData.geStateTo();

	double radius = getDocument().getPreferences().state_diameter/2.;

	CTMPoint startPoint = stateFrom.getLocation();
	CTMPoint endPoint = stateTo.getLocation();

	double dx, dy, l, ll = 0.;

	// create an AffineTransform that maps a view location to a linkView
	// location.
	CTMPoint midPoint = CTMPoint.getMidpointInstance( startPoint, endPoint );
	CTMPoint ctrlPoint = linkData.getCtrl();

	if ( stateFrom==stateTo )
	{
	  ll = ctrlPoint.distance( 0, 0 );
	  dx = -ctrlPoint.y/ll;
	  dy = ctrlPoint.x/ll;

	  // in/out points are 0.6radians left and right of ctrlPoint vector
	  mTempTransform.setTransform( dx, dy, -dy, dx, 0., 0. );
	  mTempTransform.rotate( -0.6 );
	  CTMPoint pointLeft = new CTMPoint( 0, ( int ) radius );
	  mTempTransform.transform( pointLeft, pointLeft );
	  startPoint.add( pointLeft );

	  mTempTransform.setTransform( dx, dy, -dy, dx, 0., 0. );
	  mTempTransform.rotate( 0.6 );
	  CTMPoint pointRight = new CTMPoint( 0, ( int ) radius );
	  mTempTransform.transform( pointRight, pointRight );
	  endPoint.add( pointRight );

	  dx = -1.0;
	  dy = 0.0;
	}
	else
	{
	  // vector startPoint->endPoint,
	  l = startPoint.distance( endPoint );
	  dx = ( startPoint.x-endPoint.x )/l;
	  dy = ( startPoint.y-endPoint.y )/l;
	}

	transform.setTransform( dx, dy, -dy, dx, ( double ) midPoint.x,
		( double ) midPoint.y );
	transform.transform( ctrlPoint, controlPoint );

	// move the transition component to the correct position on the curve.
	CTMPoint transitionComponentLocation = new CTMPoint( controlPoint );

	// if the curve is vertically inverted then put the transition component
	// on theother side of it.
	if ( ctrlPoint.y<0 )
	{
	  dx = -dx;
	  dy = -dy;
	}

	transitionComponentLocation.y += ( panelTransitionComponents.getHeight()+8 )
		*dx/2.;
	transitionComponentLocation.x += ( panelTransitionComponents.getWidth()+8 )
		*-dy/2.;
	transitionComponentLocation.subtract( panelTransitionComponents
		.getWidth()/2, panelTransitionComponents.getHeight()/2 );

	CTMPoint midPointTemp = CTMPoint.getMidpointInstance( startPoint,
							endPoint ).subtract( startPoint );

	midPointTemp.scale( ( stateFrom!=stateTo )?0.5: ( ll/25. ) );

	controlPoint1.set( controlPoint ).subtract( midPointTemp );
	controlPoint2.set( controlPoint ).add( midPointTemp );

	// move start point from center of state to edge in direction of control
	// point
	if ( stateFrom!=stateTo )
	{
	  l = radius/startPoint.distance( controlPoint1 );
	  dx = ( controlPoint1.x-startPoint.x )*l;
	  dy = ( controlPoint1.y-startPoint.y )*l;
	  startPoint.add( ( int ) dx, ( int ) dy );
	}

	// move end point from center of state to edge in direction of control
	// point
	l = endPoint.distance( controlPoint2 );
	dx = ( controlPoint2.x-endPoint.x )/l;
	dy = ( controlPoint2.y-endPoint.y )/l;
	if ( stateFrom!=stateTo )
	{
	  endPoint.translate( ( int ) ( radius*dx ), ( int ) ( radius*dy ) );
	}

	// Generate the arrow head coordinates
	arrowLeft.setLocation( 8, -4 );
	arrowRight.setLocation( 8, 4 );

	mTempTransform.setTransform( dx, dy, -dy, dx, ( double ) endPoint.x,
		( double ) endPoint.y );
	mTempTransform.transform( arrowLeft, arrowLeft );
	mTempTransform.transform( arrowRight, arrowRight );

	// now set the bounds correctly
	CTMRectangle tempRect = new CTMRectangle();
	Point[] tempPoints = new Point[]
						 {
						 startPoint, endPoint, controlPoint,
						 controlPoint1, controlPoint2, arrowLeft, arrowRight,
						 transitionComponentLocation};

	tempRect.add( tempPoints );
	tempRect.add( transitionComponentLocation.x
		+panelTransitionComponents.getWidth(),
		transitionComponentLocation.y
		+panelTransitionComponents.getHeight() );

	tempRect.width++;
	tempRect.height++;

	CTMPoint.translate( tempPoints, -tempRect.x, -tempRect.y );

	panelTransitionComponents.setLocation( transitionComponentLocation );

	setBounds( tempRect );

	curve1.setCurve( startPoint, controlPoint1, controlPoint );
	curve2.setCurve( controlPoint, controlPoint2, endPoint );

	getViewComponent().checkSize();
  }

  /**
   * Get the data for this link component belongs to.
   *
   * @return The data for this link component.
   */
  public CTMDataLink getLinkData()
  {
	return ( CTMDataLink ) getData();
  }
}
