package turingmachine;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.border.*;

/**
 * A component that holds a single turing machine transition. This gets added to
 * CTMComponentLinks.
 *
 * @see CTMComponentLink
 */
class CTMComponentTransition
	extends CTMComponent
{
  /**
   * Text field that holds the transition.
   */
  private JTextField textField;

  /**
   * Default constructor.
   *
   * @param dataTransition The data that this component represents.
   */
  public CTMComponentTransition( CTMDataTransition dataTransition )
  {
	super( dataTransition );

	CTMPreferences prefs = getDocument().getPreferences();
	Font font = new Font( "", Font.PLAIN, prefs.link_font_size );

	setLayout( new GridLayout( 1, 1 ) );

	textField = new JTextField( "( , )", 5 );
	textField.setHorizontalAlignment( JTextField.CENTER );
	textField.setFont( font );
	textField.setBorder( new EmptyBorder( 0, 0, 0, 0 ) );
	textField.setForeground( prefs.state_font_colour );
	textField.setOpaque( false );
	add( textField, BorderLayout.CENTER );

	textField.addMouseListener( new MouseAdapter()
	{
	  public void mouseReleased( MouseEvent e )
	  {
		alignCaretPosition();
	  }
	} );

	textField.addKeyListener( new KeyAdapter()
	{
	  public void keyTyped( KeyEvent e )
	  {}

	  public void keyPressed( KeyEvent e )
	  {
		switch ( e.getKeyCode() )
		{
		  case KeyEvent.VK_BACK_SPACE:
			int p = ( textField.getCaretPosition()<3 )?3:1;
			textField.setCaretPosition( p );
			alignCaretPosition();
			e.consume();
			break;

		  case KeyEvent.VK_DELETE:
			e.consume();
			getDataTransition().delete();
			break;

		  case KeyEvent.VK_ESCAPE:
			getViewComponent().requestFocus();
			break;

			// On enter add a new transition to the link.
		  case KeyEvent.VK_ENTER:
			getDataLink().addTransition();
			break;
		}

		alignCaretPosition();
	  }

	  public void keyReleased( KeyEvent e )
	  {
		if ( Character.isLetterOrDigit( e.getKeyChar() ) )
		{
		  int field = textField.getCaretPosition()/2-1;
		  String fieldValue = String.valueOf(
							  textField.getText().charAt(
							  textField.getCaretPosition()-1 ) )
							  .toUpperCase();
		  boolean success = false;

		  switch ( field )
		  {
			case 0:
			  success = getDataTransition().setSymbol( fieldValue );
			  break;
			case 1:
			  success = getDataTransition().setDirection( fieldValue );
			  break;
			default:
			  throw new RuntimeException(
				  "CTMComponentTransition invalid field position" );
		  }

		  // pre-save caret position as it get's lost when text
		  // changes
		  int cp = textField.getCaretPosition();
		  textField.setText( "("+getDataTransition().getMatch()
			  +","+getDataTransition().getDirection()+")" );

		  if ( success )
		  {
			int p = ( cp>=4||cp<=1 )?1:3;
			textField.setCaretPosition( p );
			textField.moveCaretPosition( p+1 );
		  }
		  else
		  {
			int p = ( cp<3 )?1:3;
			textField.setCaretPosition( p );
			textField.moveCaretPosition( p+1 );

			String errorMessage;
			if ( p==3 )
			{
			  errorMessage = "The symbol entered was not a valid symbol for this type of Machine.\r\n"
							 +"Valid symbols  are '"
							 +getDocument().getSymbols()
							 +getDocument().getDirections()+"'";
			}
			else
			{
			  errorMessage = "The symbol entered was not a valid for this Machine.\r\n"
							 +"Valid symbols are '"
							 +getDocument().getSymbols()+"'";
			}
			JOptionPane.showMessageDialog(
				CTMComponentTransition.this, errorMessage,
				"Syntax Error", JOptionPane.ERROR_MESSAGE );
		  }
		}
	  }
	} );
	textField.addFocusListener( new FocusAdapter()
	{
	  public void focusGained( FocusEvent e )
	  {
		alignCaretPosition();
	  }
	} );

	textField.setText( "         " );
	textField.setCaretPosition( 1 );
	textField.moveCaretPosition( 2 );
  }

  /**
   * Align the caret position to closest editable position.
   */
  public void alignCaretPosition()
  {
	int p = ( textField.getCaretPosition()<3 )?1:3;
	textField.setCaretPosition( p );
	textField.moveCaretPosition( p+1 );
  }

  /**
   * Select the text field.
   */
  public void select()
  {
	textField.requestFocus();
	textField.setCaretPosition( 1 );
	textField.moveCaretPosition( 2 );
  }

  /**
   * Synchronize this link with it's data component.
   */
  public void sync()
  {
	int p = ( textField.getCaretPosition()<3 )?1:3;

	textField.setText( "("+getDataTransition().getMatch()+","
		+getDataTransition().getDirection()+")" );
	textField.setCaretPosition( p );
	textField.moveCaretPosition( p+1 );

	super.sync();
  }

  /**
   * Get the dataTransition for this component.
   *
   * @return DataTransition for this component.
   */
  public CTMDataTransition getDataTransition()
  {
	return ( CTMDataTransition ) getData();
  }

  /**
   * Get the dataLink for this component. Transitions belong to links. The
   * link is the parent of the transition.
   *
   * @return DataLink for this component.
   */
  public CTMDataLink getDataLink()
  {
	return ( CTMDataLink ) getDataTransition().getDataParent();
  }
}
