package turingmachine;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

import static turingmachine.CTMFrame.TOOL_OPTION;

/**
 * Component that handles a view of a document.
 */
class CTMComponentView
	extends JComponent
{
  /**
   * The document this view is a view of.
   */
  private CTMDataDocument mDocument;

  /**
   * The editing arrow for adding links.
   */
  private CTMComponentArrow mLinkArrow;

  /**
   * The control for modifying link curves.
   */
  private CTMComponentControl mControlComponent;

  /**
   * The control for moving multiple states.
   */
  private CTMComponentDragSelectArea mDragArea;

  /**
   * Whether or not grid is showing.
   */
  private boolean isShowingGrid = false;

  /**
   * Whether to snap components to grid.
   */
  private boolean isSnapToGrid = false;

  /**
   * Grid minor tick mark spacing.
   */
  private int gridMinorSpacing = 32;

  /**
   * Cursor for creating objects.
   */
  private static Cursor cursor_createState = CTMUtility.createCursor(
	  "Resource/Cursor/createState.gif", new Point( 16, 15 ),
	  "Create State" ), cursor_createComment = CTMUtility.createCursor(
	  "Resource/Cursor/comment.gif", new Point( 1, 1 ), "Create Comment" );

  /**
   * Constructor.
   *
   * @param aDocument The document that this component is a view of.
   */
  public CTMComponentView( CTMDataDocument aDocument )
  {
	// allow components to be put at any XY position on screen
	setLayout( null );

	// set the internal reference for the parent document
	mDocument = aDocument;

	mLinkArrow = new CTMComponentArrow();
	mLinkArrow.setVisible( false );
	mLinkArrow.setEnabled( false );
	add( mLinkArrow );

	mControlComponent = new CTMComponentControl();
	mControlComponent.setVisible( false );
	add( mControlComponent );

	mDragArea = new CTMComponentDragSelectArea();
	mDragArea.setVisible( false );
	add( mDragArea );

	// listener for adding a state
	addMouseListener( new MouseAdapter()
	{
	  public void mousePressed( MouseEvent e )
	  {
		this_mousePressed( e );
	  }

	  public void mouseReleased( MouseEvent e )
	  {
		this_mouseReleased( e );
	  }
	} );

	this.addMouseMotionListener( new MouseMotionAdapter()
	{
	  public void mouseDragged( MouseEvent e )
	  {
		this_mouseDragged( e );
	  }
	} );

	setDoubleBuffered( false );
  }

  /**
   * Toggle showing grid.
   */
  public void toggleGrid()
  {
	isShowingGrid = !isShowingGrid;
	repaint();
  }

  /**
   * Returns whether or not the grid is showing.
   *
   * @return boolean true if the grid is showing.
   */
  public boolean isShowingGrid()
  {
	return isShowingGrid;
  }

  /**
   * Get grid minor divisions.
   *
   * @return int Grid minor divisions.
   */
  public int getGridMinorDivisions()
  {
	return 4;
  }

  /**
   * Get grid major spacing.
   *
   * @return int Grid major spacing.
   */
  public int getGridMajorSpacing()
  {
	return getGridMinorSpacing()*getGridMinorDivisions();
  }

  /**
   * Set grid minor spacing.
   *
   * @param value int Grid minor spacing.
   */
  public void setGridMinorSpacing( int value )
  {
	// sanity check
	if ( value<4 )
	{
	  value = 4;
	}
	this.gridMinorSpacing = value;
	repaint();
  }

  /**
   * Get grid minor spacing.
   *
   * @return Grid minor spacing.
   */
  public int getGridMinorSpacing()
  {
	return gridMinorSpacing;
  }

  /**
   * Toggle snap to grid setting.
   */
  public void toggleSnapToGrid()
  {
	isSnapToGrid = !isSnapToGrid;
	repaint();
  }

  /**
   * Return whether or not to snap to grid.
   *
   * @return boolean true snap to grid, else false.
   */
  public boolean isSnapToGrid()
  {
	return isSnapToGrid;
  }

  /**
   * Returns whether or not components should snap to grid, based on whether
   * control key is pressed and state of snap to grid.
   *
   * @param e The mouse event that caused this method to be called.
   * @return boolean true if components are to snap to grid.
   */
  public boolean isSnapToGrid( MouseEvent e )
  {
	return ( ( e.getModifiersEx()&MouseEvent.CTRL_DOWN_MASK )>0 )
		?!isSnapToGrid
		:isSnapToGrid;
  }

  /**
   * Returns whether or not the parent window is in the foreground.
   *
   * @return boolean true if the parent window is selected.
   */
  public boolean isSelected()
  {
	JComponent parent = this;

	while ( ( parent = ( JComponent ) parent.getParent() )!=null )
	{
	  if ( parent instanceof JInternalFrame )
	  {
		return ( ( JInternalFrame ) parent ).isSelected();
	  }
	}

	// should never get here
	throw new RuntimeException(
		"isSelected() component is not child of a window" );
  }

  /**
   * Paint the view.
   *
   * @param g The graphics context to paint to.
   */
  public void paint( Graphics g )
  {
	Graphics2D g2 = ( Graphics2D ) g;

	g.setColor( getDocument().getPreferences().view_backgroundColour );
	Rectangle clipBounds = g.getClipBounds();
	g.fillRect( clipBounds.x, clipBounds.y, clipBounds.width,
		clipBounds.height );

	// draw grid
	if ( isShowingGrid )
	{
	  int gridMajorSpacing = getGridMajorSpacing();
	  int gridMinorSpacing = getGridMinorSpacing();
	  Color gridMajorColor = Color.GRAY;
	  Color gridMinorColor = Color.LIGHT_GRAY;

	  g.setColor( gridMinorColor );

	  int sx = ( int ) ( clipBounds.x/gridMinorSpacing )*gridMinorSpacing;
	  for ( int x = sx; x<=clipBounds.x+clipBounds.width; x += gridMinorSpacing )
	  {
		g.setColor( x%gridMinorSpacing==x%gridMajorSpacing
			?gridMajorColor
			:gridMinorColor );
		g.drawLine( x, clipBounds.y, x, clipBounds.y
			+clipBounds.height );
	  }

	  int sy = ( int ) ( clipBounds.y/gridMinorSpacing )*gridMinorSpacing;
	  for ( int y = sy; y<=clipBounds.y+clipBounds.height; y += gridMinorSpacing )
	  {
		g.setColor( y%gridMinorSpacing==y%gridMajorSpacing
			?gridMajorColor
			:gridMinorColor );
		g.drawLine( clipBounds.x, y, clipBounds.x+clipBounds.width, y );
	  }
	}

	boolean antiAlias = getDocument().getPreferences().view_antialias;
	g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING,
		antiAlias?RenderingHints.VALUE_ANTIALIAS_ON
		:RenderingHints.VALUE_ANTIALIAS_OFF );

	paintChildren( g );
  }

  /**
   * Get the correct cursor for the current tool and cursor position.
   *
   * @return The cursor to use.
   */
  public Cursor getCursor()
  {
	switch ( getSelectedTool() )
	{
	  case CREATE_NODE:
		return cursor_createState;
	  case CREATE_COMMENT:
		return cursor_createComment;
	  default:
		return Cursor.getDefaultCursor();
	}
  }

  /**
   * Get the arrow component used for linking states together.
   *
   * @return This views arrow component.
   */
  public CTMComponentArrow getArrowComponent()
  {
	return mLinkArrow;
  }

  /**
   * Get the control component used for modifying links curvature.
   *
   * @return This views control component.
   */
  public CTMComponentControl getControlComponent()
  {
	return mControlComponent;
  }

  /**
   * Get the document this view is rendering.
   *
   * @return The document that this view represents.
   */
  public CTMDataDocument getDocument()
  {
	return mDocument;
  }

  /**
   * Performs tool on a click event.
   *
   * @param e The mouse pressed event.
   */
  private void this_mousePressed( MouseEvent e )
  {
	if ( ( e.getModifiers()&e.BUTTON1_MASK )!=0 )
	{
	  getControlComponent().setVisible( false );
	  mDragArea.setVisible( false );
	  requestFocus();

	  Point pt = e.getPoint();
	  if ( isSnapToGrid( e ) )
	  {
		int gms = getGridMinorSpacing();
		pt.x = ( ( int ) ( pt.x+gms/2 )/gms )*gms;
		pt.y = ( ( int ) ( pt.y+gms/2 )/gms )*gms;
	  }

	  if ( getSelectedTool()==CTMFrame.TOOL_OPTION.CREATE_NODE )
	  {
		mDocument.stateAdd( pt );
	  }
	  else
	  if ( getSelectedTool()==CTMFrame.TOOL_OPTION.CREATE_COMMENT )
	  {
		mDragArea.setVisible( true );
		mDragArea.setBounds( pt.x, pt.y, 1, 1 );
	  }
	}
  }

  /**
   * Performs tool move event.
   *
   * @param e The mouse pressed event.
   */
  private void this_mouseDragged( MouseEvent e )
  {
	if ( ( e.getModifiers()&e.BUTTON1_MASK )!=0 )
	{
	  if ( getSelectedTool()==CTMFrame.TOOL_OPTION.CREATE_COMMENT )
	  {
		int w = e.getPoint().x-mDragArea.getX();
		int h = e.getPoint().y-mDragArea.getY();
		mDragArea.setSize( w, h );
		checkSize();
	  }
	}
  }

  /**
   * Performs tool release event.
   *
   * @param e The mouse pressed event.
   */
  private void this_mouseReleased( MouseEvent e )
  {
	if ( ( e.getModifiers()&e.BUTTON1_MASK )!=0 )
	{
	  if ( getSelectedTool()==CTMFrame.TOOL_OPTION.CREATE_COMMENT )
	  {
		mDragArea.setVisible( false );
		mDocument.commentAdd( mDragArea.getBounds() );
	  }
	}
  }

  /**
   * Get the Component that data is referred to in.
   *
   * @param aData The data the defines a components parameters.
   * @return The component if found, else null.
   */
  public CTMComponent getComponentByData( CTMData aData )
  {
	// loop through all the views components looking for a component of type
	// CTMComponent that has a data the same as aData.
	Component c[] = ( Component[] ) getComponents();

	for ( int cnt = 0; cnt<c.length; cnt++ )
	{
	  if ( c[cnt] instanceof CTMComponent )
	  {
		CTMComponent component = ( CTMComponent ) c[cnt];
		if ( component.getData()==aData )
		{
		  return component;
		}
	  }
	}

	return null;
  }

  /**
   * Sync this view with its document.
   */
  public void sync()
  {
	/*
	 * Go through document and make sure:
	 * - there aren't any new components
	 * - current components don't need update
	 * - componenets havn't been deleted
	 */
	int mainComponentCount = 0;
	for ( Iterator i = getDocument().componentIterator(); i.hasNext(); )
	{
	  Object data = i.next();

	  if ( data instanceof CTMDataState )
	  {
		CTMDataState state = ( CTMDataState ) data;

		// get the component it pertains to and create if needed
		if ( getComponentByData( state )==null )
		{
		  CTMComponentState newState = new CTMComponentState( state );
		  add( newState, mainComponentCount );
		}

		for ( Iterator j = state.linksFromIterator(); j.hasNext(); )
		{
		  CTMDataLink link = ( CTMDataLink ) j.next();
		  CTMComponentLink newLink = ( CTMComponentLink ) getComponentByData( link );

		  // get the component it pertains to and create if needed
		  if ( newLink==null&&!link.isDeleted() )
		  {
			newLink = new CTMComponentLink( link );
			add( newLink );
			newLink.sync(); // make sure new link doesn't have to
			// sync transitions
		  }
		}
	  }
	  else if ( data instanceof CTMDataComment )
	  {
		CTMDataComment comment = ( CTMDataComment ) data;
		// get the component it pertains to and create if needed
		if ( getComponentByData( comment )==null )
		{
		  CTMComponentComment newComment = new CTMComponentComment( comment );

		  add( newComment, mainComponentCount );
		  if ( hasFocus() )
		  {
			newComment.requestFocus();
		  }
		}
	  }
	  else
	  {
		throw new RuntimeException( "Unknown component type found "+data );
	  }

	  mainComponentCount++;
	}

	syncComponents( this );
	checkSize();
  }

  /**
   * Go through all components in container and make sure:
   * - they still exist in the document?
   * - they don't need syncing
   *
   * @param container Container
   */
  public void syncComponents( Container container )
  {
	Component c[] = ( Component[] ) container.getComponents();

	for ( int cnt = 0; cnt<c.length; cnt++ )
	{
	  if ( c[cnt] instanceof CTMComponent )
	  {
		CTMComponent component = ( CTMComponent ) c[cnt];

		if ( component.requiresSync() )
		{
		  component.sync();
		}
		if ( component.getData().isDeleted() )
		{
		  container.remove( component );
		  container.repaint(); // TODO: CTMComponentLinks
		  // backgrounds don't redraw on
		  // deletion.
		}
	  }
	}
  }

  /**
   * Get the application currently selected tool.
   *
   * @return Current tool in use by user.
   */
  public CTMFrame.TOOL_OPTION getSelectedTool()
  {
	return getRootFrame().getSelectedTool();
  }

  /**
   * Get the root frame this menubar belongs to.
   *
   * @return The root frame this menubar belongs to.
   */
  private CTMFrame getRootFrame()
  {
	Component c = this;
	while ( ! ( ( c = c.getParent() )instanceof CTMFrame ) )
	{
	  ;
	}
	return ( CTMFrame ) c;
  }

  /**
   * Check the document is large enough to hold its components.
   */
  public void checkSize()
  {
	// make sure point doesn't exceed document size
	int maxX = 0;
	int minX = 65536;
	int maxY = 0;
	int minY = 65536;

	Component[] comps = getComponents();
	for ( int i = 0; i<comps.length; i++ )
	{
	  Component comp = comps[i];
	  Point pt = comp.getLocation();

	  if ( pt.x<minX )
	  {
		minX = pt.x;
	  }
	  if ( pt.y<minY )
	  {
		minY = pt.y;
	  }
	  pt.x += comp.getWidth();
	  pt.y += comp.getHeight();
	  if ( pt.x>maxX )
	  {
		maxX = pt.x;
	  }
	  if ( pt.y>maxY )
	  {
		maxY = pt.y;
	  }
	}
	Dimension newSize = new Dimension();
	newSize.setSize( maxX, maxY );

	if ( !newSize.equals( getPreferredSize() ) )
	{
	  setPreferredSize( newSize );
	  getParent().validate();
	}
	// TODO: move all components down and right, so all components are in
	// view i.e. no negative locations
  }

}
