package turingmachine;

/**
 * The class that all CTMDocument viewable elements must inherit from.
 */
abstract class CTMData
	implements java.io.Serializable
{
  /**
   * Serial id, so future version will load old files
   */
  static final long serialVersionUID = 1154131147262332537L;

  /**
   * The document this data belongs to.
   */
  private CTMDataDocument mDocument; // TODO this cannot be here

  /**
   * The count of modifications to this object.
   */
  private transient int mModificationCount = 0;

  /**
   * Whether or not this object has been deleted.
   */
  private transient boolean mDeleted = false;

  /**
   * The parent data for this data, e.g. dataTransitions parent is dataLink.
   */
  private CTMData mDataParent; // TODO this cannot be here. ??

  /**
   * Constructor.
   *
   * @param aDocument
   *            The document this data belongs to.
   */
  public CTMData( CTMDataDocument aDocument )
  {
	mDocument = aDocument;

	touch();
  }

  /**
   * Constructor.
   *
   * @param aDocument The document this data belongs to.
   * @param aDataParent Tha parent data of this data.
   */
  public CTMData( CTMDataDocument aDocument, CTMData aDataParent )
  {
	this( aDocument );
	mDataParent = aDataParent;
  }

  /**
   * Update this data's modification count to reflect some change in its
   * sub-class.
   */
  public void touch()
  {
	mModificationCount++;

	if ( mDataParent!=null )
	{
	  mDataParent.touch();
	}

	mDocument.setModified( true );
  }

  /**
   * Get this data's current modification value.
   *
   * @return The current modification value.
   */
  public int getModificationCount()
  {
	return mModificationCount;
  }

  /**
   * Set this data as being deleted.
   */
  public void delete()
  {
	mDeleted = true;
	touch();
  }

  /**
   * Has this data been deleted?
   *
   * @return true - yes, false - no.
   */
  public boolean isDeleted()
  {
	return mDeleted;
  }

  /**
   * Get the document this state belongs to.
   *
   * @return The document this state belongs to.
   */
  public CTMData getDataParent()
  {
	return mDataParent;
  }

  /**
   * Get the document this state belongs to.
   *
   * @return The document this state belongs to.
   */
  public CTMDataDocument getDocument()
  {
	return mDocument;
  }
}
