package turingmachine;

import java.awt.*;
import java.io.*;
import java.util.*;

import javax.swing.*;

/**
 * A turing machine document.
 */
class CTMDataDocument
	extends Observable implements Serializable
{
  /**
   * Serial id, so future version will load old files.
   */
  static final long serialVersionUID = 3241875946049968928L;

  /**
   * Document modification state.
   */
  private transient boolean mModified;

  /**
   * Whether or not this machine is deterministic.
   */
  private boolean mIsDeterministic = true;

  /**
   * Filename this document was loaded from.
   */
  private transient String mFileName = "";

  /**
   * Number of states created.
   */
  private int mStateCount = 0;

  /**
   * Document title.
   */
  private String mTitle = "Untitled Machine";

  /**
   * Turing machine symbols.
   */
  private String mSymbols = "01";

  /**
   * Turing machine directions.
   */
  private String mDirections = "";

  /**
   * Default prefix for new states.
   */
  private String mStatePrefix = "q";

  /**
   * List of states in document.
   */
  private ArrayList<CTMData> mComponents = new ArrayList<CTMData> ();

  /**
   * Prefereneces for document.
   */
  private CTMPreferences preferences = new CTMPreferences();

  /**
   * This applications default file filter.
   */
  private static CTMFileFilter defaultFileFilter = new CTMFileFilter(
	  "Turing Machine (.tm)", new String[]
	  {"tm"} );

  /**
   * Constructor.
   */
  public CTMDataDocument()
  {
  }

  /**
   * Gets the document title.
   *
   * @return The documents title.
   */
  public String getTitle()
  {
	return mTitle;
  }

  /**
   * Getter for documents filename.
   *
   * @return String Documents filename.
   */
  public String getFilename()
  {
	return mFileName;
  }

  /**
   * Sets the document title.
   *
   * @param aTitle The new document title.
   */
  public void setTitle( String aTitle )
  {
	// if this title isn't the same as the current one it is modified
	if ( mTitle.compareTo( aTitle )!=0 )
	{
	  // assign the title
	  mTitle = aTitle;

	  // document has changed
	  setModified( true );

	  // TODO move this somewhere
	}
  }

  /**
   * Gets the symbols for the turing machine.
   *
   * @return Symbols that machine may use.
   */
  public String getSymbols()
  {
	return mSymbols;
  }

  /**
   * Sets the symbols for the turing machine.
   *
   * @param aSymbols The allowed symbols.
   */
  public void setSymbols( String aSymbols )
  {
	// if this symbols isn't the same as the current one then it is modified
	if ( mSymbols.compareTo( aSymbols )!=0 )
	{
	  // update symbols
	  mSymbols = aSymbols;

	  // document has changed, but only if it isn't a new document
	  if ( mSymbols.length()!=0 )
	  {
		setModified( true );
	  }

	  // TODO: go through doc and make sure symbols are all valid
	}
  }

  /**
   * Gets turing machine tape direction symbols.
   *
   * @return Turing machine tape direction symbols.
   */
  public String getDirections()
  {
	return mDirections;
  }

  /**
   * Sets turing machine tape direction symbols.
   *
   * @param aDirections The valid directions for this machine.
   */
  public void setDirections( String aDirections )
  {
	if ( mDirections!=aDirections )
	{
	  // document has changed, but only if it isn't a new document
	  if ( mDirections.length()!=0 )
	  {
		setModified( true );
	  }

	  mDirections = aDirections;
	}
  }

  /**
   * Ask whether the use wants to save the document.
   *
   * @param parentFrame Frame that save dialog will belong to.
   * @return True if it is OK to close the document else false.
   */
  public boolean saveQuestion( CTMFrame parentFrame )
  {
	String[] options =
		{
		"Cancel", "Lose Changes", "Save"};

	int action = JOptionPane.showOptionDialog( parentFrame,
				 "Do you want to save '"+getTitle()+"' before closing?",
				 "Save document?", JOptionPane.YES_NO_CANCEL_OPTION,
				 JOptionPane.QUESTION_MESSAGE, null, options, options[0] );

	switch ( action )
	{
	  default: // closed dialog, cancel, unknown
		return false;

	  case 1: // don't save
		return true;

	  case 2: // save
		return save( parentFrame );
	}
  }

  /**
   * Load a document.
   *
   * @param parentFrame Frame that load dialog will belong to.
   * @return The loaded document.
   */
  public static CTMDataDocument load( CTMFrame parentFrame )
  {
	// show the file chooser dialog
	JFileChooser dlg = new JFileChooser();
	dlg.setFileFilter( defaultFileFilter );

	// if a file was choosen
	if ( dlg.showOpenDialog( parentFrame )==dlg.APPROVE_OPTION )
	{
	  try
	  {
		String filename = dlg.getSelectedFile().getPath();

		if ( parentFrame.fileIsAlreadyOpen( filename ) )
		{
		  int action = JOptionPane
					   .showConfirmDialog(
					   parentFrame,
					   "The file you have chosen is already open, do you wish to re-open a copy?",
					   "Open already open file?",
					   JOptionPane.YES_NO_OPTION,
					   JOptionPane.QUESTION_MESSAGE );

		  if ( action!=JOptionPane.YES_OPTION )
		  {
			return null;
		  }
		}

		// open the file and read in the document
		FileInputStream file = new FileInputStream( dlg
			.getSelectedFile() );
		ObjectInputStream input = new ObjectInputStream( file );
		CTMDataDocument document = ( CTMDataDocument ) input.readObject();
		document.mFileName = filename;

		return document;
	  }
	  catch ( java.lang.Exception e )
	  {
		JOptionPane
			.showMessageDialog(
			parentFrame,
			"Unable to understand format.\nThere is nothing you can do that will load that file with this version of 'Turing Machine Simulator'",
			"Error loading file.",
			JOptionPane.ERROR_MESSAGE );
	  }
	}

	return null;
  }

  /**
   * Save this document under a different name.
   *
   * @param parentFrame Frame that load dialog will belong to.
   * @return true if save succeeded else false.
   */
  public boolean saveAs( CTMFrame parentFrame )
  {
	String oldFileName = mFileName;

	// setting the documents filename to nothing will force the
	// file chooser dialog to come up.
	mFileName = "";

	// save failed?
	if ( save( parentFrame )==false )
	{
	  // set the documents filename back to the old name.
	  mFileName = oldFileName;
	  return false;
	}

	return true;
  }

  /**
   * Provide a save dialog and save the document.
   *
   * @param parentFrame Frame that load dialog will belong to.
   * @return true if save succeeded else false.
   */
  public boolean save( CTMFrame parentFrame )
  {
	if ( mFileName.length()==0 )
	{
	  JFileChooser dlg = new JFileChooser();
	  dlg.setFileFilter( defaultFileFilter );
	  if ( dlg.showSaveDialog( parentFrame )!=dlg.APPROVE_OPTION )
	  {
		return false;
	  }

	  mFileName = dlg.getSelectedFile().getPath();

	  if ( dlg.getSelectedFile().exists() )
	  {
		String[] options =
			{
			"Cancel", "Overwrite File"};

		int action = JOptionPane
					 .showOptionDialog(
					 parentFrame,
					 "The file '"
					 +mFileName
					 +"' already exists, do you want to overwrite it?",
					 "Overwrite file?", JOptionPane.YES_NO_OPTION,
					 JOptionPane.QUESTION_MESSAGE, null, options,
					 options[0] );

		switch ( action )
		{
		  default: // closed dialog, cancel
			return false;

		  case 1: // Overwrite
			break;
		}
	  }
	}

	ObjectOutputStream out = null;

	try
	{
	  FileOutputStream file = new FileOutputStream( mFileName );
	  out = new ObjectOutputStream( file );
	  out.writeObject( this );
	  out.flush();
	  setModified( false );
	  return true;
	}
	catch ( java.io.IOException IOE )
	{
	  JOptionPane.showMessageDialog( parentFrame, IOE.getMessage(),
		  "Error in saving file", JOptionPane.ERROR_MESSAGE );
	}
	finally
	{
	  try
	  {
		if ( out!=null )
		{
		  out.close();
		}
	  }
	  catch ( java.io.IOException ex )
	  {
		ex.printStackTrace();
	  }
	}

	return false;
  }

  /**
   * Method for getting the components iterator as mComponents is private.
   *
   * @return An Iterator for the states in the document.
   */
  public Iterator<CTMData> componentIterator()
  {
	return mComponents.iterator();
  }

  /**
   * Add a state to the document and update views.
   *
   * @param aPoint Where to put the state.
   */
  public void stateAdd( Point aPoint )
  {
	CTMDataState newState = new CTMDataState( this, mStatePrefix
							+mStateCount++, aPoint );

	// If this is the first state to be placed make it a start state
	if ( mStateCount==1 )
	{
	  newState.setType( CTMDataState.START_NODE );
	}
	mComponents.add( newState );

	notifyObservers();
  }

  /**
   * Override notifyObservers to set changed value.
   */
  public void notifyObservers()
  {
	setChanged();
	super.notifyObservers();
  }

  /**
   * Add a comment to the document and update views.
   *
   * @param bounds Where to put the comment.
   */
  public void commentAdd( Rectangle bounds )
  {
	CTMDataComment newComment = new CTMDataComment( this, "", bounds );
	mComponents.add( newComment );
	notifyObservers();
  }

  /**
   * Bring a component to the front.
   *
   * @param component The component to reorder.
   */
  public void componentToFront( CTMData component )
  {
	// reorder state in document to front
	mComponents.remove( component );
	mComponents.add( component );

	setModified( true );
  }

  /**
   * Send a component to the back.
   *
   * @param component The component to reorder.
   */
  public void componentToBack( CTMData component )
  {
	// reorder state in document to front
	mComponents.remove( component );
	mComponents.add( mComponents.size(), component );

	setModified( true );
  }

  /**
   * Check document is a valid program.
   *
   * @param parentFrame Frame that load dialog will belong to.
   * @return boolean true document is fine, else false.
   */
  public boolean validate( CTMFrame parentFrame )
  {
	// 1. check that all links from states have different input conditions
	// for deterministic TM
	// 2. and all input conditions are valid
	// 3. all output conditions are valid
	// 4. check there is an end state, and a start state as well
	// 5. and end-state has no from links
	// TODO: check there is a route from start state to end state
	// TODO: check there is a route to all states
	// TODO: configurable error options for 1,2,3,4
	// TODO: make a dialog listing all errors with double click going to
	// error
	boolean hasEndState = false, hasStartState = false;

	for ( CTMData data:mComponents )
	{
	  if ( data instanceof CTMDataState )
	  {
		CTMDataState dataState = ( CTMDataState ) data;

		if ( dataState.getType()==CTMDataState.START_NODE )
		{
		  if ( hasStartState )
		  {
			JOptionPane.showMessageDialog( parentFrame,
				"Turing machine has multiple start states.",
				"Invalid Machine", JOptionPane.ERROR_MESSAGE );
			return false;
		  }
		  else
		  {
			hasStartState = true;
		  }
		}
		if ( dataState.getType()==CTMDataState.END_NODE )
		{
		  if ( hasEndState )
		  {
			JOptionPane.showMessageDialog( parentFrame,
				"Turing machine has multiple end states.",
				"Invalid Machine", JOptionPane.ERROR_MESSAGE );
			return false;
		  }
		  else
		  {
			if ( dataState.hasLinksFrom() )
			{
			  JOptionPane.showMessageDialog( parentFrame, "End state '"
				  +dataState.getName()
				  +"' has transitions from it.",
				  "Invalid Machine",
				  JOptionPane.ERROR_MESSAGE );
			  return false;
			}
			hasEndState = true;
		  }
		}

		for ( Iterator j = dataState.linksFromIterator(); j.hasNext(); )
		{
		  CTMDataLink dataLink = ( CTMDataLink ) j.next();

		  // used to check state is deterministic
		  String foundMatches = "";

		  for ( Iterator k = dataLink.transitionsIterator(); k
							 .hasNext(); )
		  {
			CTMDataTransition dataTransition = ( CTMDataTransition ) k
				.next();

			if ( getSymbols().indexOf( dataTransition.getMatch() )==-1 )
			{
			  JOptionPane.showMessageDialog( parentFrame,
				  "State transition '"+dataState.getName()
				  +"'->'"
				  +dataLink.geStateTo().getName()
				  +"' has invalid match symbol '"
				  +dataTransition.getMatch()+"'.",
				  "Invalid Machine",
				  JOptionPane.ERROR_MESSAGE );
			  return false;
			}

			if ( isDeterministic()&&foundMatches.indexOf( dataTransition.getMatch() )!=-1 )
			{
			  JOptionPane.showMessageDialog( parentFrame,
				  "State transition '"+dataState.getName()
				  +"'->'"
				  +dataLink.geStateTo().getName()
				  +"' is non-deterministic.",
				  "Invalid Machine",
				  JOptionPane.ERROR_MESSAGE );
			  return false;
			}

			foundMatches += dataTransition.getMatch();
		  }
		}
	  }
	}

	if ( !hasStartState )
	{
	  JOptionPane.showMessageDialog( parentFrame,
		  "Turing machine has no start state.", "Invalid Machine",
		  JOptionPane.ERROR_MESSAGE );
	  return false;
	}

	if ( !hasEndState )
	{
	  JOptionPane.showMessageDialog( parentFrame, "Turing machine has no end state.",
		  "Invalid Machine", JOptionPane.ERROR_MESSAGE );
	  return false;
	}

	JOptionPane.showMessageDialog( parentFrame, "Turing machine is valid." );

	return true;
  }

  /**
   * Set the status of the documents modified flag.
   *
   * @param aModified Modification state to set.
   */
  public void setModified( boolean aModified )
  {
	mModified = aModified;
	notifyObservers();
  }

  /**
   * Get whether or not this document has been modified.
   *
   * @return true - it has been modified, else false.
   */
  public boolean isModified()
  {
	return mModified;
  }

  /**
   * Set whether or not this machine is determininistic.
   *
   * @param isDeterministic true if it is deterministic, else false.
   */
  public void setIsDeterministic( boolean isDeterministic )
  {
	this.mIsDeterministic = isDeterministic;
	setModified( true );
  }

  /**
   * Get whether or not this machine is determininistic.
   *
   * @return true if it is deterministic, else false.
   */
  public boolean isDeterministic()
  {
	return mIsDeterministic;
  }

  /**
   * Gets document preferences.
   *
   * @return CTMPreferences
   */
  public CTMPreferences getPreferences()
  {
	return preferences;
  }
}
