package turingmachine;

import java.awt.geom.*;
import java.io.*;
import java.util.*;

/**
 * Data for a link between states.
 */
class CTMDataLink
	extends CTMData implements Serializable
{
  /**
   * Serial id, so future version will load old files
   */
  static final long serialVersionUID = -2055624832118373633L;

  /**
   * The state this link links from.
   */
  private CTMDataState mStateFrom;

  /**
   * The state this link links to.
   */
  private CTMDataState mStateTo;

  /**
   * The control point for the curve on this link.
   */
  private CTMPoint mCtrl;

  /**
   * The transitions for this link.
   */
  private ArrayList<CTMDataTransition> mTransitions;

  /**
   * Constructor.
   *
   * @param document The document this state belongs to.
   * @param stateFrom The state this links from.
   * @param stateTo The state this links to.
   */
  public CTMDataLink( CTMDataDocument document, CTMDataState stateFrom,
	  CTMDataState stateTo )
  {
	super( document );

	mTransitions = new ArrayList<CTMDataTransition> ();

	this.mStateFrom = stateFrom;
	this.mStateTo = stateTo;

	// default control position
	CTMPreferences prefs = document.getPreferences();

	if ( mStateFrom==mStateTo )
	{
	  mCtrl = new CTMPoint( 0, prefs.link_controlOffset_monoconnect );
	}
	else
	{
	  mCtrl = new CTMPoint( 0, prefs.link_controlOffset_biconnect );
	}
  }

  /**
   * Add new transition data to this link data.
   */
  public void addTransition()
  {
	CTMDataDocument doc = getDocument();

	// cannot have more links than symbols
	if ( mTransitions.size()<doc.getSymbols().length() )
	{
	  CTMDataTransition transition = new CTMDataTransition( doc, this );
	  mTransitions.add( transition );
	  touch();
	}
  }

  /**
   * Handle notification from a data transition that it has been deleted.
   *
   * @param dataTransition The data transition that has been deleted.
   */
  public void transitionDeleted( CTMDataTransition dataTransition )
  {
	mTransitions.remove( dataTransition );

	// Delete the link if that was the last transition on the link.
	if ( mTransitions.size()==0 )
	{
	  delete();
	}
  }

  /**
   * Get the transition data iterator.
   *
   * @return The transition data iterator.
   */
  public Iterator<CTMDataTransition> transitionsIterator()
  {
	return mTransitions.iterator();
  }

  /**
   * Get the current control position.
   *
   * @return The location of the control position is fromState to toState
   *         midpoint coordinates.
   */
  public CTMPoint getCtrl()
  {
	return new CTMPoint( mCtrl );
  }

  /**
   * Set the control points position in fromState to toState midpoint
   * coordinates.
   *
   * @param aPoint The new control point location.
   */
  public void setCtrl( Point2D aPoint )
  {
	// set the control point
	mCtrl.setLocation( aPoint );

	// touch the state so the view will update on next repaint
	touch();
  }

  /**
   * Get the state that this link links from.
   *
   * @return The state this link links from.
   */
  public CTMDataState getStateFrom()
  {
	return mStateFrom;
  }

  /**
   * Get the state that this link links to.
   *
   * @return The state this link links to.
   */
  public CTMDataState geStateTo()
  {
	return mStateTo;
  }

  /**
   * Delete this link and any transition that it contains.
   */
  public void delete()
  {
	super.delete();

	while ( mTransitions.size()!=0 )
	{
	  CTMDataTransition dataTransition = mTransitions.get( 0 );
	  dataTransition.delete();
	  mTransitions.remove( dataTransition );
	}

	// remove link from the target states from list
	mStateFrom.linkRemove( this );
  }
}
