package turingmachine;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import java.util.*;

/**
 * State data class.
 */
class CTMDataState
	extends CTMData implements Serializable
{
  /**
   * Serial id, so future version will load old files
   */
  static final long serialVersionUID = -2330377904366754385L;

  /**
   * The name of this states displayed state.
   */
  private String mName;

  /**
   * This states position in the document.
   */
  private CTMPoint mLocation;

  /**
   * is the breakpoint set on this state.
   */
  private boolean mBreakPoint;

  /**
   * links from this state that go to another.
   */
  private ArrayList<CTMDataLink> mLinksTo;

  /**
   * links from other states that come to this ( used for faster updates )
   * maybe make transient and autogen on load.
   */
  private ArrayList<CTMDataLink> mLinksFrom;

  /**
   * What type of state is this?
   */
  private int mType;

  /**
   * Start state type.
   */
  public final static int START_NODE = 0;

  /**
   * Normal state type.
   */
  public final static int NORMAL_NODE = 1;

  /**
   * End state type.
   */
  public final static int END_NODE = 2;

  /**
   * Gosub state type.
   */
  public final static int GOSUB_NODE = 3;

  /**
   * Constructor.
   *
   * @param aDocument The document this state data is part of.
   * @param aStateName The name of this states state.
   * @param aPoint The initial location of the state.
   */
  public CTMDataState( CTMDataDocument aDocument, String aStateName, Point aPoint )
  {
	super( aDocument );

	mName = aStateName;
	mLocation = new CTMPoint( aPoint );
	mType = NORMAL_NODE;
	mBreakPoint = false;

	mLinksTo = new ArrayList<CTMDataLink> ();
	mLinksFrom = new ArrayList<CTMDataLink> ();
  }

  /**
   * Add a link from this state to another state. If a link already exists
   * returns the existing link.
   *
   * @param toState The State to link to.
   * @return The newly created link data.
   */
  public CTMDataLink linkAdd( CTMDataState toState )
  {
	CTMDataLink newLink = getLinkToState( toState );
	if ( newLink==null )
	{
	  newLink = new CTMDataLink( getDocument(), this, toState );
	  mLinksFrom.add( newLink );
	  toState.linkAddTo( newLink );
	}

	return newLink;
  }

  /**
   * Add a reference to another states link to this state.
   *
   * @param aFromLink The other states linkTo instance.
   */
  private void linkAddTo( CTMDataLink aFromLink )
  {
	mLinksTo.add( aFromLink );
  }

  /**
   * Remove a link from this state.
   *
   * @param aLink The link to remove.
   */
  public void linkRemove( CTMDataLink aLink )
  {
	// remove from internal list
	mLinksFrom.remove( aLink );

	// remove from external states to list
	aLink.geStateTo().linkRemoveTo( aLink );

	// touch so that it will update on view update
	touch();
  }

  /**
   * Remove a reference from another states link to this state.
   *
   * @param aLink The other states link instance remove.
   */
  private void linkRemoveTo( CTMDataLink aLink )
  {
	mLinksTo.remove( aLink );
  }

  /**
   * Is the breakpoint set.
   *
   * @return true - yes, false - no.
   */
  public boolean isBreakPointSet()
  {
	return mBreakPoint;
  }

  /**
   * Sets the breakpoint.
   *
   * @param breakpoint The value to set the breakpoint to.
   */
  public void setBreakPoint( boolean breakpoint )
  {
	// set breakpoint
	this.mBreakPoint = breakpoint;

	// touch so that view components will update
	touch();
  }

  /**
   * Get the state type.
   *
   * @return This states type.
   */
  public int getType()
  {
	return mType;
  }

  /**
   * Set the states type.
   *
   * @param aType The type of state to set it to.
   */
  public void setType( int aType )
  {
	// check aType is valid
	if ( aType<START_NODE||aType>GOSUB_NODE )
	{
	  throw new RuntimeException( "invalid state type exception.\n"
		  +aType+" is not a valid value" );
	}

	// set type
	mType = aType;

	// touch so that view components will update
	touch();
  }

  /**
   * Get the states center point location.
   *
   * @return The center point location of the state.
   */
  public CTMPoint getLocation()
  {
	return new CTMPoint( mLocation );
  }

  /**
   * Set this states position within the document.
   *
   * @param aPoint The new location for the state.
   */
  public void setLocation( Point2D aPoint )
  {
	// if the position is different from original then document is modified
	if ( !aPoint.equals( mLocation ) )
	{
	  CTMDataDocument doc = getDocument();
	  mLocation.setLocation( aPoint );

	  // touch this so its' components will sync
	  touch();

	  // set all links as modified so they will sync
	  touchLinks();
	}
  }

  /**
   * Move this state relative to its' current position.
   *
   * @param aX The horizontal shift offset.
   * @param aY The vertical shift offset.
   */
  public void translate( int aX, int aY )
  {
	mLocation.x += aX;
	mLocation.y += aY;

	touchLinks();
	touch();
  }

  /**
   * Set all sub links as modified so they will sync.
   */
  private void touchLinks()
  {
	// touch links from this state
	for ( CTMDataLink linkFrom:mLinksFrom )
	{
	  linkFrom.touch();
	}

	// touch links to this state
	for ( CTMDataLink linkTo:mLinksTo )
	{
	  linkTo.touch();
	}
  }

  /**
   * Get the state name.
   *
   * @return This states state name.
   */
  public String getName()
  {
	return mName;
  }

  /**
   * Set the state name.
   *
   * @param name The new name for the state.
   */
  public void setStateName( String name )
  {
	// is the statename different from what it already is?
	if ( name.compareTo( mName )!=0 )
	{
	  // set the state name
	  mName = name;

	  // touch data so that view components will update
	  touch();
	}
  }

  /**
   * Bring this state to the front in the document.
   */
  public void toFront()
  {
	getDocument().componentToFront( this );
  }

  /**
   * Delete this state from its document
   */
  public void delete()
  {
	super.delete();

	// clear out all links coming from this state
	// view will be updated by state delete on viewComponent
	while ( mLinksFrom.size()!=0 )
	{
	  mLinksFrom.get( 0 ).delete();
	}

	// clear out all links coming to this state
	// view will be updated by state delete on viewComponent
	while ( mLinksTo.size()!=0 )
	{
	  mLinksTo.get( 0 ).delete();
	}
  }

  /**
   * Check whether this state links from this to aState ( used as check before
   * creating links ).
   *
   * @param aState
   *            the state to check to.
   * @return true - it does link to aState, else false.
   */
  public CTMDataLink getLinkToState( CTMDataState aState )
  {
	for ( CTMDataLink linkFrom:mLinksFrom )
	{
	  if ( linkFrom.geStateTo()==aState )
	  {
		return linkFrom;
	  }
	}

	return null;
  }

  /**
   * Get the iterator to this states links To.
   *
   * @return An iterator for this states links to.
   */
  public Iterator<CTMDataLink> linksToIterator()
  {
	return mLinksTo.iterator();
  }

  /**
   * Get the iterator to this states links From.
   *
   * @return An iterator for this states links from.
   */
  public Iterator<CTMDataLink> linksFromIterator()
  {
	return mLinksFrom.iterator();
  }

  /**
   * Get whether or not this state has links leaving it.
   *
   * @return true if state has links leaving it, else false.
   */
  public boolean hasLinksFrom()
  {
	return mLinksFrom.size()!=0;
  }
}
